#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
01-protest-classifier.py
Purpose: fine-tune ResNet18 to predict images of on-street protests. Replication
           of the Casas & Webb-Williams' Images As Data paper in which we 
           originally fine tuned AlexNet using Caffe. 
@author: Andreu Casas (acs706@nyu.edu)
"""

# MODULES
#==============================================================================
from __future__ import print_function, division
import torch
import torch.nn as nn
import torch.optim as optim
from torch.optim import lr_scheduler
import torchvision
from torchvision import datasets, models
import matplotlib.pyplot as plt
plt.ion()   # interactive mode
import os
# - DEVELOPMENT: move to the directory where the local copy of vpgtools is
# - PRODUCTION: no need to move to that directory
os.chdir('/Users/andreucasas/Desktop/repos/')
import vpgtools.vpgtools as vpg

# PATHS & CONSTANTS
#==============================================================================
# - path to the directory where the train and test images are. The directory
#       need to have the following structure:
# (e.g.)
# -- protest_exp
#   |
#    -- train
#      |
#       -- negative
#      | 
#       -- positive
#   |
#    -- test
#      |
#       -- negative
#      | 
#       -- positive
DATA_DIR = '/Users/andreucasas/Desktop/repos/vpgtools/data/protest_exp/'

# DATA WRANGLING
#==============================================================================
# - specifying the transformations to apply to the images
preprocess = vpg.data_tranform()

# - specifying where the train and test images are
image_datasets = {x: datasets.ImageFolder(os.path.join(DATA_DIR, x),
                                          preprocess[x])
                  for x in ['train', 'test']}

# - creating an object to easily load the train and test images
dataloaders = {x: torch.utils.data.DataLoader(image_datasets[x], batch_size=4,
                                             shuffle=True)
              for x in ['train', 'test']}

# - getting some extra information such as the size of the dataset and the
#       class number
dataset_sizes = {x: len(image_datasets[x]) for x in ['train', 'test']}
class_names = image_datasets['train'].classes

# - checking if the machine has GPUs and enabling GPU computing if true
use_gpu = torch.cuda.is_available()


# DATA EXPLORATION
#==============================================================================
# - taking a quick look at some examples
# - get a batch of training data
inputs, classes = next(iter(dataloaders['train']))

# - make a grid from batch
out = torchvision.utils.make_grid(inputs)
vpg.imshow(out, title=[class_names[x] for x in classes])


# MAIN: Fine Tuning the Binary Classifier
#==============================================================================
# - load a pretrained model: ResNet18
model = models.resnet18(pretrained=True)

# - check number of features in last fully connected layer
num_ftrs = model.fc.in_features 

# - change the output layer from X (e.g. 1000) classes to a binary outcome
model.fc = nn.Linear(num_ftrs, 2) # For ResNet18, last step is 512 -> 2

# - specify some model hyperparameters
criterion = nn.CrossEntropyLoss() # loss function
optimizer = optim.SGD(model.parameters(), 
                         lr=0.001, 
                         momentum=0.9) # how to perform optimization
exp_lr_scheduler = lr_scheduler.StepLR(optimizer, 
                                       step_size=7, 
                                       gamma=0.1) # how to update learning rate

# - specify whether to use gpu 
if use_gpu:
    model = model.cuda()
    
# - being training
model_ft = vpg.finetune_binary_classifier(image_datasets = image_datasets, 
                                          dataloaders = dataloaders, 
                                          use_gpu = use_gpu,
                                          model = model,
                                          criterion = criterion,
                                          optimizer = optimizer, 
                                          scheduler = exp_lr_scheduler, 
                                          num_epochs = 4)

# Visualize the model: see 4 predictions at a time
vpg.visualize_model(model = model_ft,
                    class_names = class_names,
                    num_images = 4,
                    dataloaders = dataloaders,
                    use_gpu = use_gpu)

# OUTPUT: save the model
torch.save(model_ft.state_dict(), 
           '/Users/andreucasas/Desktop/repos/vpgmodels/protest_model.pth.tar')
